#include "OgreVisibilityTerrainSceneManager.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>


namespace Ogre {

//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::VisibilityTerrainSceneManager(
	GtpVisibility::VisibilityManager *visManager): 
mVisibilityManager(visManager), 
mIsDepthPass(false),
mShowVisualization(false),
mRenderNodesForViz(false),
mRenderNodesContentForViz(false),
mVisualizeCulledNodes(false),
mSkipTransparents(false),
mDelayRenderTransparents(true),
mUseDepthPass(false)
{
	mHierarchyInterface = new OctreeHierarchyInterface(this, mDestRenderSystem);

	//mDisplayNodes = true;
	//mShowBoundingBoxes = true;

	// TODO: set maxdepth to reasonable value
	mMaxDepth = 50;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::InitDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		// Init
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
        mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::~VisibilityTerrainSceneManager()
{
	if (mHierarchyInterface)
	{
		delete mHierarchyInterface;
		mHierarchyInterface = NULL;
	}
}
//-----------------------------------------------------------------------
/*Pass *VisibilityTerrainSceneManager::setPass(Pass* pass)
{
	static bool myLastUsedVertexProgram = false;
    static bool myLastUsedFragmentProgram = false;

	// set depth fill pass only if depth write enabled
	Pass *usedPass = pass;

	if (mIsDepthPass && pass->getDepthWriteEnabled())
	{
		usedPass = mDepthPass;		
	}

    if (mIlluminationStage == IRS_RENDER_TO_TEXTURE)
    {
        // Derive a special shadow caster pass from this one
        usedPass = deriveShadowCasterPass(usedPass);
    }
    else if (mIlluminationStage == IRS_RENDER_MODULATIVE_PASS)
    {
        usedPass = deriveShadowReceiverPass(usedPass);
    }

	bool passSurfaceAndLightParams = true;

	// take original pass here
    if (usedPass->hasVertexProgram())
    {
        mDestRenderSystem->bindGpuProgram(usedPass->getVertexProgram()->_getBindingDelegate());
        // bind parameters later since they can be per-object
        myLastUsedVertexProgram = true;
        // does the vertex program want surface and light params passed to rendersystem?
        passSurfaceAndLightParams = usedPass->getVertexProgram()->getPassSurfaceAndLightStates();
    }
    else
    {
        // Unbind program?
        if (myLastUsedVertexProgram)
        {
            mDestRenderSystem->unbindGpuProgram(GPT_VERTEX_PROGRAM);
            myLastUsedVertexProgram = false;
        }
        // Set fixed-function vertex parameters
    }

    if (passSurfaceAndLightParams)
    {
        // Set surface reflectance properties, only valid if lighting is enabled
        if (usedPass->getLightingEnabled())
        {
            mDestRenderSystem->_setSurfaceParams( 
                usedPass->getAmbient(), 
                usedPass->getDiffuse(), 
                usedPass->getSpecular(), 
                usedPass->getSelfIllumination(), 
                usedPass->getShininess(),
				usedPass->getVertexColourTracking() );
        }

        // Dynamic lighting enabled?
        mDestRenderSystem->setLightingEnabled(usedPass->getLightingEnabled());
    }

    // Using a fragment program?
    if (usedPass->hasFragmentProgram())
    {
        mDestRenderSystem->bindGpuProgram(
            usedPass->getFragmentProgram()->_getBindingDelegate());
        // bind parameters later since they can be per-object
        myLastUsedFragmentProgram = true;
    }
    else
    {
        // Unbind program?
        if (myLastUsedFragmentProgram)
        {
            mDestRenderSystem->unbindGpuProgram(GPT_FRAGMENT_PROGRAM);
            myLastUsedFragmentProgram = false;
        }

        // Set fixed-function fragment settings

        // Fog (assumes we want pixel fog which is the usual)
        // New fog params can either be from scene or from material
        FogMode newFogMode;
        ColourValue newFogColour;
        Real newFogStart, newFogEnd, newFogDensity;
        if (usedPass->getFogOverride())
        {
            // New fog params from material
            newFogMode = usedPass->getFogMode();
            newFogColour = usedPass->getFogColour();
            newFogStart = usedPass->getFogStart();
            newFogEnd = usedPass->getFogEnd();
            newFogDensity = usedPass->getFogDensity();
        }
        else
        {
            // New fog params from scene
            newFogMode = mFogMode;
            newFogColour = mFogColour;
            newFogStart = mFogStart;
            newFogEnd = mFogEnd;
            newFogDensity = mFogDensity;
        }
        mDestRenderSystem->_setFog(newFogMode, newFogColour, newFogDensity, newFogStart, newFogEnd);

	}

    // The rest of the settings are the same no matter whether we use programs or not

    // Set scene blending
    mDestRenderSystem->_setSceneBlending(
        usedPass->getSourceBlendFactor(), usedPass->getDestBlendFactor());


    // Texture unit settings

    Pass::TextureUnitStateIterator texIter =  usedPass->getTextureUnitStateIterator();
    size_t unit = 0;
    while(texIter.hasMoreElements())
    {
        TextureUnitState* pTex = texIter.getNext();
        mDestRenderSystem->_setTextureUnitSettings(unit, *pTex);
        ++unit;
    }
    // Disable remaining texture units
    mDestRenderSystem->_disableTextureUnitsFrom(usedPass->getNumTextureUnitStates());

    // Set up non-texture related material settings
    // Depth buffer settings
    mDestRenderSystem->_setDepthBufferFunction(usedPass->getDepthFunction());
    mDestRenderSystem->_setDepthBufferCheckEnabled(usedPass->getDepthCheckEnabled());
    mDestRenderSystem->_setDepthBufferWriteEnabled(usedPass->getDepthWriteEnabled());
    mDestRenderSystem->_setDepthBias(usedPass->getDepthBias());
	// Alpha-reject settings
	mDestRenderSystem->_setAlphaRejectSettings(
		usedPass->getAlphaRejectFunction(), usedPass->getAlphaRejectValue());
    // Set colour write mode
    // Right now we only use on/off, not per-channel
    bool colWrite = usedPass->getColourWriteEnabled();
    mDestRenderSystem->_setColourBufferWriteEnabled(colWrite, colWrite, colWrite, colWrite);
    // Culling mode
    mDestRenderSystem->_setCullingMode(usedPass->getCullingMode());
    // Shading
    mDestRenderSystem->setShadingType(usedPass->getShadingMode());

	return usedPass;
}*/
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::ShowVisualization(Camera *cam)
{
	// add player camera for visualization purpose
	try {
		Camera *c;
		if ((c = getCamera("PlayerCam")) != NULL)
		{
			getRenderQueue()->addRenderable(c);
		}   
    }
    catch(...)
    {
        // ignore
    }
	for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
	{
		getRenderQueue()->addRenderable(*it);
	}
	if (mRenderNodesForViz || mRenderNodesContentForViz)
	{
		// change node material so it is better suited for visualization
		MaterialPtr nodeMat = MaterialManager::getSingleton().getByName("Core/NodeMaterial");
		nodeMat->setAmbient(1, 1, 0);
		nodeMat->setLightingEnabled(true);
		nodeMat->getTechnique(0)->getPass(0)->removeAllTextureUnitStates();

		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			
			if (mRenderNodesForViz)
			{
				getRenderQueue()->addRenderable(*it);
				// addbounding boxes instead of node itself
				//(*it)->_addBoundingBoxToQueue(getRenderQueue());
			}
			if (mRenderNodesContentForViz) 
			{
				(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
			}
		}
	}
	
}
//-----------------------------------------------------------------------
Pass *VisibilityTerrainSceneManager::setPass(Pass* pass)
{
	// setting vertex program is not efficient
	//Pass *usedPass = ((mIsDepthPass && !pass->hasVertexProgram()) ? mDepthPass : pass);		
	// set depth fill pass only if depth write enabled
	Pass *usedPass = (mIsDepthPass && pass->getDepthWriteEnabled() ? mDepthPass : pass);
	
	if (mIsDepthPass)
	{
		// set vertex program of current pass
        if (pass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram(pass->getVertexProgramName());

			if (mDepthPass->hasVertexProgram())
			{
				const GpuProgramPtr& prg = mDepthPass->getVertexProgram();
				// Load this program if not done already
				if (!prg->isLoaded())
					prg->load();
				// Copy params
				mDepthPass->setVertexProgramParameters(pass->getVertexProgramParameters());
			}
		}
		else if (mDepthPass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram("");
		}
	}
	/*else if (mIsItemBufferPass)
	{
		usedPass = mItemBufferPass;
	}*/
	SceneManager::setPass(usedPass);

	return usedPass;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
{
	//-- show visible scene nodes and octree bounding boxes from last frame
	if (mShowVisualization)
    {
		ShowVisualization(cam);
	}
	else 
	{
		mVisible.clear();
	    mBoxes.clear();
		
		// if there is no depth pass =>
		// we interleave identification and rendering of objects 
		// in _renderVisibibleObjects

		// only shadow casters will be rendered in shadow texture pass
		mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_renderVisibleObjects()
{
	// create material for depth pass
	InitDepthPass();

	// visualization: apply standard rendering
	if (mShowVisualization)
	{	
		TerrainSceneManager::_renderVisibleObjects();
		return;
	}


	//-- hierarchical culling
	// the objects of different layers (e.g., background, scene, 
	// overlay) must be identified and rendered one after another

	bool leaveTransparentsInQueue = mDelayRenderTransparents && !mUseDepthPass;

	// possible two cameras (one for culling, one for rendering)
	mHierarchyInterface->InitFrame(mOctree, mCameraInProgress, 
							mCullCamera ? getCamera("CullCamera") : NULL,
							leaveTransparentsInQueue);

	// call initframe to reset culling manager stats
	mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);
	
	mSkipTransparents = false;

	//-- render background, in case there is one
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
	addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);
	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);

	SceneManager::_renderVisibleObjects();


#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	_deleteRenderedQueueGroups(false);
#endif

	//-- render visible objects (i.e., all but overlay and skies late)
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_LATE);
	addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);

	// transparents are skipped from hierarchical rendering
	// => they need sorting, thus we render them afterwards
	mSkipTransparents = mDelayRenderTransparents;

	// set state for depth pass
    mIsDepthPass = mUseDepthPass;
	
	/** 
	  * the hierarchical culling algorithm
	  * for depth pass: will just find objects and update depth buffer
	  * for delayed rendering: will render all but transparents
	**/
	
	mVisibilityManager->ApplyVisibilityCulling();


	// for depth pass: add visible nodes found with the visibility culling
	if (mUseDepthPass)
	{
		/*for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			(*it)->_addToRenderQueue(mCameraInProgress, getRenderQueue(), false);
		}*/
		mIsDepthPass = false;
	}

	mSkipTransparents = false;

	//-- now we can render all remaining queue objects
	// for depth pass: all
	// for delayed rendering: transparents, overlay 
	clearSpecialCaseRenderQueues();
	SceneManager::_renderVisibleObjects();

	//WriteLog(); // write out stats
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_updateSceneGraph(Camera* cam)
{
	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
    mHierarchyInterface->SetNumOctreeNodes(mNumOctreeNodes);
#endif
	TerrainSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "ShowVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesContentForViz")
	{
		mRenderNodesContentForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || TerrainSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctreeNodes;
		return true;
	}
	
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && TerrainSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionValues(const String & key, StringVector &refValueList)
{
	return TerrainSceneManager::getOptionValues( key, refValueList);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionKeys(StringVector & refKeys)
{
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys(refKeys) || TerrainSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::setVisibilityManager(GtpVisibility::VisibilityManager *visManager)
{
	mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityTerrainSceneManager::getVisibilityManager( void )
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Depth pass: " << StringConverter::toString(mUseDepthPass) << ", "
	  << "Delay transparents: " << StringConverter::toString(mDelayRenderTransparents) << ", "
	  << "Use optimization: " << StringConverter::toString(mHierarchyInterface->GetUseOptimization()) << ", "
	  << "Algorithm type: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctreeNodes << ", " 
	  << "Traversed nodes: " << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::renderObjects(const RenderPriorityGroup::TransparentRenderablePassList& objs, 
            bool doLightIteration, const LightList* manualLightList)
{
	if (!mSkipTransparents)
	{
		OctreeSceneManager::renderObjects(objs, doLightIteration, manualLightList);
	}
}

} // namespace Ogre
